@@ -11,10 +11,12 @@ import asyncio
 import copy
 import functools
 import io
+import itertools
 import logging
 import math
 import os
 import sys
+from collections import defaultdict
 from concurrent.futures import Executor
 from dataclasses import dataclass
 from functools import reduce
@@ -25,16 +27,20 @@ import torch
 from torch.distributed._shard.sharded_tensor import (
     Shard as ShardedTensorShard,
     ShardedTensor,
-    ShardedTensorMetadata,
     ShardMetadata,
 )
 from torch.distributed._shard.sharding_spec import ChunkShardingSpec
+from torch.distributed._shard.sharding_spec._internals import (
+    _check_shard_metadata_pair_overlap,
+)
 
 from .io_types import BufferConsumer, BufferStager, BufferType, ReadReq, WriteReq
+from .knobs import get_max_chunk_size_bytes, get_max_shard_size_bytes
 from .manifest import (
     ChunkedTensorEntry,
     Entry,
     ObjectEntry,
+    PrimitiveEntry,
     Shard,
     ShardedTensorEntry,
     TensorEntry,
@@ -51,12 +57,6 @@ from .serialization import (
     torch_load_from_bytes,
     torch_save_as_bytes,
 )
-from .torch_dist_checkpoint.metadata import (
-    ExtendedTensorMetadata,
-    StorageMetadata,
-    TensorReadRequest,
-)
-from .torch_dist_checkpoint.resharding import prepare_sharded_tensor_read
 
 logger: logging.Logger = logging.getLogger(__name__)
 
@@ -69,21 +69,20 @@ class Chunk:
 
 
 class ChunkedTensorIOPreparer:
-    DEFAULT_MAX_CHUNK_SIZE_BYTES: int = 512 * 1024 * 1024
-
     @staticmethod
     def chunk_tensor(
         tensor: torch.Tensor,
         chunking_dim: int = 0,
+        chunk_sz_bytes: Optional[int] = None,
     ) -> List[Chunk]:
+        chunk_sz_bytes = chunk_sz_bytes or get_max_chunk_size_bytes()
+
         # for 0-d case, reshape to 1-d
         if tensor.ndim == 0:
             tensor = tensor.view(-1)
 
         tensor_sz_bytes = tensor.numel() * tensor.element_size()
-        n_chunks = math.ceil(
-            tensor_sz_bytes / ChunkedTensorIOPreparer.DEFAULT_MAX_CHUNK_SIZE_BYTES
-        )
+        n_chunks = math.ceil(tensor_sz_bytes / chunk_sz_bytes)
         tensor_chunks = torch.chunk(tensor, chunks=n_chunks, dim=chunking_dim)
 
         curr_offsets = [0] * tensor.ndim
@@ -117,6 +116,7 @@ class ChunkedTensorIOPreparer:
         storage_path: str,
         tensor: torch.Tensor,
         chunking_instruction: List[Chunk],
+        is_async_snapshot: bool = False,
         _tensor_prepare_func: Optional[
             Callable[[torch.Tensor, bool], torch.Tensor]
         ] = None,
@@ -126,8 +126,10 @@ class ChunkedTensorIOPreparer:
         for chunk in chunking_instruction:
             suffix = "_".join(str(x) for x in chunk.offsets)
             chunk_entry, chunk_write_reqs = TensorIOPreparer.prepare_write(
-                f"{storage_path}_{suffix}",
-                cls._get_subtensor_view(tensor, chunk),
+                storage_path=f"{storage_path}_{suffix}",
+                tensor=cls._get_subtensor_view(tensor, chunk),
+                is_async_snapshot=is_async_snapshot,
+                _tensor_prepare_func=_tensor_prepare_func,
             )
             chunks.append(
                 Shard(offsets=chunk.offsets, sizes=chunk.sizes, tensor=chunk_entry)
@@ -163,8 +165,6 @@ class ChunkedTensorIOPreparer:
 
 
 class ShardedTensorIOPreparer:
-    DEFAULT_MAX_SHARD_SIZE_BYTES: int = 512 * 1024 * 1024
-
     @staticmethod
     def subdivide_shard(
         shard: torch.Tensor,
@@ -197,11 +197,61 @@ class ShardedTensorIOPreparer:
             subdivided.append((sub_view, sub_offsets, sub_sizes))
         return subdivided
 
+    @staticmethod
+    def _shards_get_overlap_region_wrt_saved_tensor(
+        saved_shard: ShardMetadata, current_shard: ShardMetadata
+    ) -> List[Tuple[int, int, int, int]]:
+        """
+        Return the overlapping region between saved_shard and current_shard.
+        There returned list has the same number of elements as the tensor's dimension.
+        For each element, we produce a tuple with the following contents:
+            (dimension, `saved_shard` offset, `current_shard` offset, length)
+
+        Offsets are relative to each shard.
+        """
+        # TODO: This is copied from
+        # torch.distributed._shard.checkpoint.resharding._shards_get_overlap_region_wrt_saved_tensor
+        # which is not in PyTorch 1.11 yet. Remove this method and import it
+        # from PyTorch directly once we drop support for 1.11.
+        narrows = []
+        for dim, (
+            saved_shard_offset,
+            current_shard_offset,
+            saved_shard_size,
+            current_shard_size,
+        ) in enumerate(
+            zip(
+                saved_shard.shard_offsets,
+                current_shard.shard_offsets,
+                saved_shard.shard_sizes,
+                current_shard.shard_sizes,
+            )
+        ):
+            min_range_end = min(
+                saved_shard_offset + saved_shard_size,
+                current_shard_offset + current_shard_size,
+            )
+
+            length = min_range_end - max(current_shard_offset, saved_shard_offset)
+
+            if saved_shard_offset > current_shard_offset:
+                offset_for_saved_tensor = 0
+                offset_for_current_tensor = saved_shard_offset - current_shard_offset
+            else:
+                offset_for_saved_tensor = current_shard_offset - saved_shard_offset
+                offset_for_current_tensor = 0
+
+            narrows.append(
+                (dim, offset_for_saved_tensor, offset_for_current_tensor, length)
+            )
+        return narrows
+
     @classmethod
     def prepare_write(
         cls,
         storage_path: str,
         obj: ShardedTensor,
+        is_async_snapshot: bool = False,
         _tensor_prepare_func: Optional[
             Callable[[torch.Tensor, bool], torch.Tensor]
         ] = None,
@@ -214,13 +264,12 @@ class ShardedTensorIOPreparer:
                 sharding_dim = sharding_spec.dim
             else:
                 sharding_dim = 0
-
             subdivided = cls.subdivide_shard(
                 shard=shard.tensor,
                 offsets=shard.metadata.shard_offsets,
                 sizes=shard.metadata.shard_sizes,
                 dim=sharding_dim,
-                max_shard_sz_bytes=cls.DEFAULT_MAX_SHARD_SIZE_BYTES,
+                max_shard_sz_bytes=get_max_shard_size_bytes(),
             )
 
             for tensor, offsets, sizes in subdivided:
@@ -228,6 +277,7 @@ class ShardedTensorIOPreparer:
                 entry, tensor_write_reqs = TensorIOPreparer.prepare_write(
                     storage_path=f"{storage_path}_{suffix}",
                     tensor=tensor,
+                    is_async_snapshot=is_async_snapshot,
                     _tensor_prepare_func=_tensor_prepare_func,
                 )
                 write_reqs += tensor_write_reqs
@@ -258,9 +308,9 @@ class ShardedTensorIOPreparer:
         else:
             out_shape = list(obj_out.shape)
         if out_shape != global_shape:
-            logger.warn(
-                "The shape of obj_out ({out_shape}) is different from the "
-                "shape of the persisted sharded tensor ({global_shape}). "
+            logger.warning(
+                f"The shape of obj_out ({out_shape}) is different from the "
+                f"shape of the persisted sharded tensor ({global_shape}). "
                 "Only the overlapping part will be loaded. "
             )
 
@@ -278,96 +328,158 @@ class ShardedTensorIOPreparer:
 
         global_shape = cls._get_global_shape(entry=entry)
         cls._validate_shape(global_shape=global_shape, obj_out=obj_out)
-        if isinstance(obj_out, ShardedTensor):
-            local_shards_out = obj_out.local_shards()
-        else:
-            local_shards_out = [
+
+        if type(obj_out) == ShardedTensor:
+            local_shards = obj_out.local_shards()
+        elif type(obj_out) == torch.Tensor:
+            local_shards = [
                 ShardedTensorShard(
                     tensor=obj_out,
                     metadata=ShardMetadata(
                         shard_offsets=[0] * len(obj_out.shape),
                         shard_sizes=list(obj_out.shape),
-                        placement="cpu",  # Unused for load
+                        placement=str(obj_out.device),
                     ),
                 )
             ]
+        else:
+            raise RuntimeError(
+                f"obj_out must either be a Tensor or ShardedTensor (got {type(obj_out)})"
+            )
 
-        metadata = ExtendedTensorMetadata(
-            tensor_metadata=ShardedTensorMetadata(),  # Unused for load
-            storage_metadata=[],
-        )
-        for shard in entry.shards:
-            metadata.storage_metadata.append(
-                StorageMetadata(
-                    shard_metadata=ShardMetadata(
-                        shard_offsets=shard.offsets,
-                        shard_sizes=shard.sizes,
-                        placement="cpu",  # Unused for load
+        # For each persisted shard, find all its overlapping regions with the
+        # local shards
+        path_byte_range_to_overlapping_regions = defaultdict(list)
+        for local_shard, shard in itertools.product(local_shards, entry.shards):
+            shard_md = ShardMetadata(
+                shard_offsets=shard.offsets,
+                shard_sizes=shard.sizes,
+                placement="cpu",
+            )
+            if not _check_shard_metadata_pair_overlap(local_shard.metadata, shard_md):
+                continue
+            path_byte_range = (shard.tensor.location, shard.tensor.byte_range_tuple)
+            path_byte_range_to_overlapping_regions[path_byte_range].append(
+                _OverlappingRegion(
+                    dst_tensor=local_shard.tensor,
+                    overlap_region=cls._shards_get_overlap_region_wrt_saved_tensor(
+                        saved_shard=shard_md,
+                        current_shard=local_shard.metadata,
                     ),
-                    storage_key=shard.tensor.location,
-                    length=0,  # Unused for load
-                    offset=0,  # Unused for load
                 )
             )
-        tensor_read_reqs = prepare_sharded_tensor_read(
-            metadata=metadata, local_shards_out=local_shards_out
-        )
-        locations_to_load = {twr.storage_key for twr in tensor_read_reqs}
 
+        # Read each persisted shard once and use it to load all its overlapping
+        # regions with the local shards
         read_reqs = []
         for shard in entry.shards:
-            if shard.tensor.location not in locations_to_load:
+            path_byte_range = (shard.tensor.location, shard.tensor.byte_range_tuple)
+            if path_byte_range not in path_byte_range_to_overlapping_regions:
                 continue
             read_reqs.append(
                 ReadReq(
                     path=shard.tensor.location,
                     buffer_consumer=ShardedTensorBufferConsumer(
-                        tensor_read_reqs=[
-                            trr
-                            for trr in tensor_read_reqs
-                            if trr.storage_key == shard.tensor.location
+                        overlapping_regions=path_byte_range_to_overlapping_regions[
+                            path_byte_range
                         ],
                         entry=shard.tensor,
                     ),
+                    byte_range=shard.tensor.byte_range_tuple,
                 )
             )
         return read_reqs
 
 
+@dataclass
+class _OverlappingRegion:
+    dst_tensor: torch.Tensor
+    overlap_region: List[
+        Tuple[int, int, int, int]
+    ]  # (dim, src_offset, dst_offset, length)
+
+    def get_views(self, src_tensor: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
+        src_view = src_tensor
+        dst_view = self.dst_tensor
+        for dim, src_offset, dst_offset, length in self.overlap_region:
+            src_view = torch.narrow(src_view, dim, src_offset, length)
+            dst_view = torch.narrow(dst_view, dim, dst_offset, length)
+        return src_view, dst_view
+
+
+def _q_params_equal(lhs: torch.Tensor, rhs: torch.Tensor) -> bool:
+    if lhs.qscheme() != rhs.qscheme():
+        return False
+    if lhs.qscheme() == torch.per_tensor_affine:
+        return (
+            lhs.q_scale() == rhs.q_scale() and lhs.q_zero_point() == rhs.q_zero_point()
+        )
+    elif lhs.qscheme() == torch.per_channel_affine:
+        return torch.allclose(
+            lhs.q_per_channel_scales(), lhs.q_per_channel_scales()
+        ) and torch.allclose(
+            lhs.q_per_channel_zero_points(), lhs.q_per_channel_zero_points()
+        )
+    else:
+        raise RuntimeError(f"Unrecognized qscheme {lhs.qscheme()}")
+
+
 @torch.jit.script
-def tensor_copy(dst: torch.Tensor, src: torch.Tensor) -> None:
+def _tensor_copy(dst: torch.Tensor, src: torch.Tensor) -> None:
     dst.detach().copy_(src)  # pragma: no cover
 
 
+@torch.jit.script
+def _tensor_dequantize(tensor: torch.Tensor) -> torch.Tensor:
+    return tensor.dequantize()  # pragma: no cover
+
+
+def tensor_copy(dst: torch.Tensor, src: torch.Tensor) -> None:
+    # When both dst and src are quantized tensors, .copy_() works as follows:
+    # - src storage -> dst storage
+    # - src qscheme -> dst qscheme
+    #
+    # We need to take special care when dst is a view of a larger tensor and
+    # the quantization schemes of src and dst are different, because the new
+    # qscheme will only be reflected on dest, which is a view object, but not
+    # on the original tensor. Directly calling .copy_() in this case will cause
+    # a region of the larger tensor's storage contain data that does not match
+    # the larger tensor's qscheme.
+
+    if src.is_quantized and (
+        not dst.is_quantized  # Copying from quantized Tensor to non-quantized Tensor is not allowed
+        or dst.qscheme() != src.qscheme()  # Quantized copy only works with same qscheme
+        or dst.dtype != src.dtype  # Quantized copy requires matching dtypes
+        or (dst._is_view() and not _q_params_equal(dst, src))  # See the top comment
+    ):
+        # TODO: tile the dequantize -> copy to reduce memory footprint
+        src = _tensor_dequantize(src)
+    _tensor_copy(dst, src)
+
+
 class ShardedTensorBufferConsumer(BufferConsumer):
     def __init__(
         self,
-        tensor_read_reqs: List[TensorReadRequest],
+        overlapping_regions: List[_OverlappingRegion],
         entry: TensorEntry,
     ) -> None:
-        self.tensor_read_reqs = tensor_read_reqs
+        self.overlapping_regions = overlapping_regions
         self.entry = entry
 
     async def consume_buffer(
         self, buf: bytes, executor: Optional[Executor] = None
     ) -> None:
-        view_to_copy = TensorBufferConsumer.deserialize_tensor(
+        deserialized = TensorBufferConsumer.deserialize_tensor(
             buf=buf, entry=self.entry
         )
-        for req in self.tensor_read_reqs:
-            for dim, (start, length) in enumerate(zip(req.offsets, req.lengths)):
-                view_to_copy = torch.narrow(view_to_copy, dim, start, length)
-
-            assert (
-                view_to_copy.size() == req.tensor.size()
-            ), f"The {req.storage_key} src/dst size does not match."
-
+        for overlapping_region in self.overlapping_regions:
+            src_view, dst_view = overlapping_region.get_views(src_tensor=deserialized)
             if executor is not None:
                 await asyncio.get_running_loop().run_in_executor(
-                    executor, tensor_copy, req.tensor, view_to_copy
+                    executor, tensor_copy, dst_view, src_view
                 )
             else:
-                req.tensor.copy_(view_to_copy)
+                tensor_copy(dst_view, src_view)
 
     def get_consuming_cost_bytes(self) -> int:
         tensor_sz_bytes = TensorIOPreparer.get_tensor_size_from_entry(self.entry)
@@ -390,16 +502,22 @@ class TensorBufferStager(BufferStager):
         self,
         tensor: torch.Tensor,
         entry: TensorEntry,
+        is_async_snapshot: bool,
         _tensor_prepare_func: Callable[[torch.Tensor, bool], torch.Tensor],
     ) -> None:
         self.tensor = tensor
         self.entry = entry
+        self.is_async_snapshot = is_async_snapshot
         self._tensor_prepare_func = _tensor_prepare_func
 
     async def stage_buffer(self, executor: Optional[Executor] = None) -> BufferType:
-        # TODO: if the custom prepared tensor is different from the original
+        is_tensor_custom_prepared = False
+        tensor = self._tensor_prepare_func(self.tensor, False)  # tracing=False
+        # If the custom prepared tensor is different from the original
         # tensor and is a CPU tensor, don't copy it.
-        self.tensor = self._tensor_prepare_func(self.tensor, False)  # tracing=False
+        if tensor.storage() != self.tensor.storage():
+            is_tensor_custom_prepared = True
+
         if self.tensor.is_cuda:
             # It would be nice to copy from GPU via DMA. However, it is very
             # difficult to figure out the safe amount of page-locked memory
@@ -411,11 +529,7 @@ class TensorBufferStager(BufferStager):
                 )
             else:
                 cpu_tensor = self.tensor.detach().to("cpu")
-        elif (
-            self.tensor.nelement() != self.tensor.storage().size()
-            or self.entry.serializer == Serializer.BUFFER_PROTOCOL
-        ):
-            # Avoid saving the entire storage when saving a view
+        elif not is_tensor_custom_prepared and self._should_copy_cpu_tensor():
             cpu_tensor = self.tensor.detach().clone()
         else:
             cpu_tensor = self.tensor.detach()
@@ -437,6 +551,32 @@ class TensorBufferStager(BufferStager):
         else:
             raise ValueError(f"Unrecognized serializer: {self.entry.serializer}.")
 
+    def _should_copy_cpu_tensor(self) -> bool:
+        if self.entry.serializer == Serializer.BUFFER_PROTOCOL and (
+            self.is_async_snapshot or not self.tensor.is_contiguous()
+        ):
+            # During async snapshot, it's not safe to use
+            # tensor_as_memoryview() on the original CPU tensor, as its data
+            # may change before the snapshot complete. Thus we make a copy.
+            #
+            # If the CPU tensor is non-contiguous, tensor_as_memoryview() will
+            # make a copy. We might as well do it here.
+            return True
+        if (
+            self.entry.serializer == Serializer.TORCH_SAVE
+            and self.tensor.nelement() > self.tensor.storage().size() * 0.8
+        ):
+            # When saving a tensor view, torch.save() saves the underlying
+            # storage backing the view. When the underlying storage is far
+            # larger than the view, it makes sense to make a copy of the view
+            # before saving it.
+            #
+            # TODO: when the view maps to a contiguous region within its
+            # storage, we can create a new tensor from the slicing the untyped
+            # storage to avoid the copy.
+            return True
+        return False
+
 
 class TensorBufferConsumer(BufferConsumer):
     def __init__(
@@ -468,7 +608,7 @@ class TensorBufferConsumer(BufferConsumer):
                 executor, tensor_copy, self.tensor, loaded
             )
         else:
-            self.tensor.copy_(loaded)
+            tensor_copy(self.tensor, loaded)
 
     def get_consuming_cost_bytes(self) -> int:
         tensor_sz_bytes = TensorIOPreparer.get_tensor_size_from_entry(self.entry)
@@ -492,6 +632,7 @@ class TensorIOPreparer:
     def prepare_write(
         storage_path: str,
         tensor: torch.Tensor,
+        is_async_snapshot: bool = False,
         _tensor_prepare_func: Optional[
             Callable[[torch.Tensor, bool], torch.Tensor]
         ] = None,
@@ -522,6 +663,7 @@ class TensorIOPreparer:
         buffer_stager = TensorBufferStager(
             tensor=tensor,
             entry=entry,
+            is_async_snapshot=is_async_snapshot,
             _tensor_prepare_func=_tensor_prepare_func,
         )
         return entry, [WriteReq(path=storage_path, buffer_stager=buffer_stager)]
@@ -551,7 +693,13 @@ class TensorIOPreparer:
                 tensor=tensor_out,
                 entry=entry,
             )
-            return [ReadReq(path=entry.location, buffer_consumer=buffer_consumer)]
+            return [
+                ReadReq(
+                    path=entry.location,
+                    byte_range=entry.byte_range_tuple,
+                    buffer_consumer=buffer_consumer,
+                )
+            ]
 
         num_chunks = math.ceil(
             cls.get_tensor_size_from_entry(entry) / buffer_size_limit_bytes
@@ -580,10 +728,21 @@ class TensorIOPreparer:
                     replicated=entry.replicated,
                 ),
             )
+            byte_range = entry.byte_range
+            if byte_range is None:
+                byte_range = (
+                    offset,
+                    offset + chunk_sz_bytes,
+                )
+            else:
+                byte_range = (
+                    byte_range[0] + offset,
+                    byte_range[0] + offset + chunk_sz_bytes,
+                )
             read_reqs.append(
                 ReadReq(
                     path=entry.location,
-                    byte_range=(offset, offset + chunk_sz_bytes),
+                    byte_range=byte_range,
                     buffer_consumer=buffer_consumer,
                 )
             )
@@ -671,11 +830,26 @@ def get_storage_path(obj: Any, logical_path: str, rank: int, replicated: bool) -
         return os.path.join(str(rank), logical_path)
 
 
+class PrimitivePreparer:
+    @staticmethod
+    def should_inline(obj: Any) -> bool:
+        type_name = type(obj).__name__
+        if type_name not in PrimitiveEntry.supported_types():
+            return False
+        # TODO for long str/bytes, return False to fall back to ObjectEntry
+        return True
+
+    @staticmethod
+    def prepare_write(obj: Any) -> PrimitiveEntry:
+        return PrimitiveEntry.from_object(obj)
+
+
 def prepare_write(
     obj: Any,
     logical_path: str,
     rank: int,
     replicated: bool,
+    is_async_snapshot: bool = False,
     _tensor_prepare_func: Optional[Callable[[torch.Tensor, bool], torch.Tensor]] = None,
 ) -> Tuple[Entry, List[WriteReq]]:
     """
@@ -691,15 +865,36 @@ def prepare_write(
         The class::`Entry` describing the object, and a list of
         class::`WriteReq` for persisting the object.
     """
+    if PrimitivePreparer.should_inline(obj):
+        entry = PrimitivePreparer.prepare_write(obj)
+        entry.replicated = replicated
+        return entry, []
+
     storage_path = get_storage_path(obj, logical_path, rank, replicated)
     if isinstance(obj, ShardedTensor):
         return ShardedTensorIOPreparer.prepare_write(
-            storage_path, obj, _tensor_prepare_func
+            storage_path=storage_path,
+            obj=obj,
+            is_async_snapshot=is_async_snapshot,
+            _tensor_prepare_func=_tensor_prepare_func,
         )
     elif isinstance(obj, torch.Tensor):
-        entry, obj_write_req = TensorIOPreparer.prepare_write(
-            storage_path, obj, _tensor_prepare_func
-        )
+        chunking_instruction = ChunkedTensorIOPreparer.chunk_tensor(obj)
+        if len(chunking_instruction) > 1:
+            entry, obj_write_req = ChunkedTensorIOPreparer.prepare_write(
+                storage_path=storage_path,
+                tensor=obj,
+                chunking_instruction=chunking_instruction,
+                is_async_snapshot=is_async_snapshot,
+                _tensor_prepare_func=_tensor_prepare_func,
+            )
+        else:
+            entry, obj_write_req = TensorIOPreparer.prepare_write(
+                storage_path=storage_path,
+                tensor=obj,
+                is_async_snapshot=is_async_snapshot,
+                _tensor_prepare_func=_tensor_prepare_func,
+            )
     else:
         entry, obj_write_req = ObjectIOPreparer.prepare_write(storage_path, obj)
 
@@ -739,5 +934,8 @@ def prepare_read(
         )
     elif isinstance(entry, ObjectEntry):
         return ObjectIOPreparer.prepare_read(entry, obj_out)
+    elif isinstance(entry, PrimitiveEntry):
+        # primitive types are stored inline in snapshot metadata
+        return []
     else:
         raise Exception(f"Unsupported entry type: {entry} ({entry.type}).")
